%%%-------------------------------------------------------------------
%%% @author Chen Slepher <slepheric@gmail.com>
%%% @copyright (C) 2017, Chen Slepher
%%% @doc
%%%
%%% @end
%%% Created : 18 Sep 2017 by Chen Slepher <slepheric@gmail.com>
%%%-------------------------------------------------------------------
-module(astranaut_traverse).

%% API
-export([traverse_fun_return/1, traverse_error/1]).
-export([map/2, map/3]).
-export([map_with_state/3, map_with_state/4]).
-export([reduce/3, reduce/4]).
-export([mapfold/3, mapfold/4]).
-export([map_m/3, map_m_children/3, m_subtrees/3, update_opts/1]).
-export([map_traverse_return/2, map_traverse_return_e/2, map_traverse_fun_return/2,
         traverse_fun_return_struct/1]).
-export([transform_mapfold_f/2]).
-export([format_error/1, parse_transform_return/2]).
-export([fun_return_to_monad/2, fun_return_to_monad/3]).
-export([monad_to_traverse_fun_return/1, monad_to_traverse_fun_return/2]).
-export([monad_deep_sequence_m/2, monad_deep_r_sequence_m/2]).

-define(TRAVERSE_FUN_RETURN, astranaut_traverse_fun_return).
-define(TRAVERSE_ERROR, astranaut_traverse_error).

-type traverse_node() :: tuple().
-type traverse_error() :: term().
-type traverse_state() :: term().
-type traverse_opts() :: #{traverse => traverse_style(), parse_transform => boolean(),
                           node => node_type(), formatter => module() }.

-type traverse_map_m_opts() :: #{traverse => traverse_style(), parse_transform => boolean(),
                                 node => node_type(), attribute => atom(), 
                                 formatter => module(),
                                 monad_class => module(), monad => term()
                                }.
-type node_type() :: attribute | pattern | expression | guard | form.
-type traverse_style() :: traverse_step() | all.
-type traverse_step() :: pre | post | leaf.
-type traverse_attr() :: #{step := traverse_step(), node := node_type()}.
-type traverse_fun_return() :: #{'__struct__' := astranaut_traverse_fun_return, 
                                 node => traverse_node(), state => traverse_state(),
                                 continue => boolean(),
                                 error => traverse_error(), warning => traverse_error(),
                                 errors => [traverse_error()], warings => [traverse_error()]} |
                               {error, traverse_error()} | {warning, traverse_node(), traverse_error()} |
                               continue | {traverse_node(), traverse_state()} | traverse_node().

-type traverse_return(ReturnType) :: ReturnType | 
                                     {ok, ReturnType, traverse_return_error(), traverse_return_error()} |
                                     {error, traverse_return_error(), traverse_return_error()}.

-type traverse_final_return(ReturnType) :: traverse_return(ReturnType) | parse_transform_return(ReturnType).

-type parse_transform_return(ReturnType) :: ReturnType |  {warning, ReturnType, parse_transform_return_error()} |
                                            {error, parse_transform_return_error(), parse_transform_return_error()}.

-type traverse_fun() :: fun((traverse_node(), traverse_attr()) -> traverse_fun_return()).
-type traverse_state_fun() :: fun((traverse_node(), traverse_state(), traverse_attr()) -> traverse_fun_return()).
-type parse_transform_module() :: module().
-type compile_file() :: string().
-type line() :: integer().
-type traverse_return_error() :: [{line(), parse_transform_module(), traverse_error()}].
-type parse_transform_return_error() :: [{compile_file(), traverse_return_error()}].

%%%===================================================================
%%% API
%%%===================================================================
traverse_fun_return(#{} = Map) ->
    Map#{'__struct__' => ?TRAVERSE_FUN_RETURN}.

traverse_error(#{} = Map) ->
    Map#{'__struct__' => ?TRAVERSE_ERROR}.

-spec map(traverse_fun(), Node) -> traverse_final_return(Node).
map(F, TopNode) ->
    map(F, TopNode, #{}).

-spec reduce(traverse_state_fun(), State, _Node) -> traverse_return(State).
reduce(F, Init, TopNode) ->
    reduce(F, Init, TopNode, #{}).

-spec map_with_state(traverse_state_fun(), State, Node) -> traverse_final_return({Node, State}).
map_with_state(F, Init, Form) ->
    map_with_state(F, Init, Form, #{}).

-spec mapfold(traverse_state_fun(), State, Node) -> traverse_return({Node, State}).
mapfold(F, Init, Node) ->
    mapfold(F, Init, Node, #{}).

-spec map(traverse_fun(), Node, traverse_opts()) -> traverse_final_return(Node).
map(F, TopNode, Opts) ->
    NF = fun(Node, State, Attr) ->
                 WalkReturn = F(Node, Attr),
                 map_walk_return(fun(NNode) -> {NNode, State} end, WalkReturn)
         end,
    map_with_state(NF, ok, TopNode, Opts).

-spec reduce(traverse_state_fun(), State, _Node, traverse_opts()) -> traverse_return(State).
reduce(F, Init, TopNode, Opts) ->
    NF = fun(Node, State, Attr) ->
                 WalkReturn = F(Node, State, Attr),
                 map_walk_return(fun(NState) -> {Node, NState} end, WalkReturn)
         end,
    map_traverse_return(
      fun({_NNode, State}) ->
              State
      end, mapfold(NF, Init, TopNode, Opts)).

-spec map_with_state(traverse_state_fun(), State, Node, traverse_opts()) -> traverse_final_return({Node, State}).
map_with_state(F, Init, Node, Opts) ->
    Reply = mapfold(F, Init, Node, Opts),
    NReply = map_traverse_return(
               fun({NNode, _State}) ->
                       NNode
               end, Reply),
    to_parse_transform_return(NReply, Node, Opts).

-spec mapfold(traverse_state_fun(), State, Node, traverse_opts()) -> traverse_return({Node, State}).
mapfold(F, Init, Node, Opts) ->
    SimplifyReturn = maps:get(simplify_return, Opts, true),
    Return = mapfold_1(F, Init, Node, Opts),
    simplify_return(Return, SimplifyReturn).

-spec map_traverse_return(fun((A) -> B), traverse_return(A)) -> parse_transform_return(B).
map_traverse_return(F, {ok, Reply, Errors, Warnings}) ->
    {ok, F(Reply), Errors, Warnings};
map_traverse_return(_F, {error, Errors, Warnings}) ->
    {error, Errors, Warnings};
map_traverse_return(F, {warning, Reply, Warnings}) ->
    {warning, F(Reply), Warnings};
map_traverse_return(F, Reply) ->
    F(Reply).

map_traverse_return_e(F, {ok, Reply, Errors, Warnings}) ->
    F(Reply, Errors, Warnings);
map_traverse_return_e(F, {error, Errors, Warnings}) ->
    F(error, Errors, Warnings);
map_traverse_return_e(F, {warning, Reply, Warnings}) ->
    F(Reply, [], Warnings);
map_traverse_return_e(F, Reply) ->
    F(Reply, [], []).


map_traverse_fun_return(F, {warning, Node, Reason}) ->
    {warning, F(Node), Reason};
map_traverse_fun_return(F, {error, Node, Reason}) ->
    {error, F(Node), Reason};
map_traverse_fun_return(_F, {error, Reason}) ->
    {error, Reason};
map_traverse_fun_return(_F, continue) ->
    continue;
map_traverse_fun_return(F, #{'__struct__' := astranaut_traverse_fun_return, node := Node} = Struct) ->
    Struct#{node => F(Node)};
map_traverse_fun_return(F, Node) ->
    F(Node).

traverse_fun_return_struct({warning, Node, Reason}) ->
    traverse_fun_return(#{warning => Reason, node => Node});
traverse_fun_return_struct({error, Node, Reason}) ->
    traverse_fun_return(#{error => Reason, node => Node});
traverse_fun_return_struct({error, Reason}) ->
    traverse_fun_return(#{error => Reason});
traverse_fun_return_struct(continue) ->
    traverse_fun_return(#{continue => true});
traverse_fun_return_struct(#{'__struct__' := ?TRAVERSE_FUN_RETURN} = Struct) ->
    Struct;
traverse_fun_return_struct(Node) ->
    traverse_fun_return(#{node => Node}).

update_opts(Opts) ->
    maps:merge(#{formatter => ?MODULE, traverse => all,
                 monad_class => astranaut_monad, 
                 monad => astranaut_traverse_monad}, Opts).

-spec map_m(traverse_fun(), Node, traverse_map_m_opts()) -> astranaut_monad:monadic(M, Node) when M :: astranaut_monad:monad().
map_m(F, Nodes, Opts) ->
    NOpts = update_opts(Opts),
    NF = transform_f(F, NOpts),
    map_m_1(NF, Nodes, NOpts).

format_error(Message) ->
    case io_lib:deep_char_list(Message) of
        true -> Message;
        _    -> io_lib:write(Message)
    end.

parse_transform_return({ok, Reply, [], []}, _File) ->
    Reply;
parse_transform_return({ok, Reply, [], Warnings}, File) ->
    {warning, Reply, [{File, Warnings}]};
parse_transform_return({ok, _Reply, Errors, Warnings}, File) ->
    {error, [{File, Errors}], [{File, Warnings}]};
parse_transform_return({error, Errors, Warnings}, File) ->
    {error, [{File, Errors}], [{File, Warnings}]};
parse_transform_return(Reply, _File) ->
    Reply.

%%====================================================================
%% Internal functions
%%====================================================================
mapfold_1(F, Init, Node, Opts) ->
    Monad = astranaut_traverse_monad:new(),
    NOpts = Opts#{monad => Monad, monad_class => astranaut_monad},
    NF = transform_mapfold_f(F, NOpts),
    NodeM = map_m(NF, Node, NOpts),
    astranaut_traverse_monad:run(NodeM, Init).

map_m_1(F, Nodes, Opts) when is_list(Nodes) ->
    monad_map_m(
      fun(Subtree) ->
              map_m_1(F, Subtree, Opts)
      end, Nodes, Opts);
map_m_1(F, NodeA, #{children := true} = Opts) ->
    Opts1 = maps:remove(children, Opts),
    map_m_children(F, NodeA, Opts1);
map_m_1(F, NodeA, #{} = Opts) ->
    SyntaxLib = syntax_lib(Opts),
    map_m_tree(F, NodeA, Opts, SyntaxLib).

map_m_tree(F, NodeA, Opts, SyntaxLib) ->
    Attr = maps:with([parent], Opts),
    NodeType = SyntaxLib:node_type(NodeA, Opts),
    PreType = 
        case SyntaxLib:subtrees(NodeA, Opts) of
            [] ->
                leaf;
            _Subtrees ->
                pre
        end,
    bind_with_continue(
      NodeA, 
      F(NodeA, Attr#{step => PreType, node => NodeType}),
      fun(NodeB) ->
              case SyntaxLib:subtrees(NodeB, Opts) of
                  [] ->
                      monad_return(NodeB, Opts);
                  Subtrees ->
                      Parent = SyntaxLib:type(NodeB),
                      monad_bind(
                        map_m_children(F, NodeB, Subtrees, Opts#{parent => Parent}, SyntaxLib),
                        fun(NodeC) ->
                                bind_with_continue(
                                  NodeC, 
                                  F(NodeC, Attr#{step => post, node => NodeType}),
                                  fun(NodeD) ->
                                          monad_return(NodeD, Opts)
                                  end, Opts)
                        end, Opts)
              end
      end, Opts).
    
bind_with_continue(NodeA, MNodeB, BMC, Opts) ->
    monad_bind(
      MNodeB,
      fun(continue) ->
              monad_return(NodeA, Opts);
         ({continue, NodeB}) ->
              monad_return(NodeB, Opts);
         (NodeB) ->
              BMC(NodeB)
      end, Opts).

map_m_children(F, Node, Opts) ->
    SyntaxLib = syntax_lib(Opts),
    Subtrees = SyntaxLib:subtrees(Node, Opts),
    map_m_children(F, Node, Subtrees, Opts, SyntaxLib).

map_m_children(_F, Node, [], Opts, _SyntaxLib) ->
    monad_return(Node, Opts);

map_m_children(F, Node, Subtrees, Opts, SyntaxLib) ->
    SyntaxType = SyntaxLib:type(Node),
    monad_bind(
      map_m_subtrees(F, Subtrees, Opts#{parent => SyntaxType}),
      fun(Subtrees1) ->
              Node1 = SyntaxLib:update_subtrees(Node, Subtrees1),
              monad_return(Node1, Opts)
      end, Opts).

syntax_lib(#{syntax_lib := SyntaxLib}) ->
    SyntaxLib;
syntax_lib(#{}) ->
    astranaut_erl_syntax.

map_m_subtrees(F, Nodes, #{sequence_children := Sequence} = Opts) ->
    Opts1 = maps:remove(sequence_children, Opts),
    SubtreesM = m_subtrees(F, Nodes, Opts1),
    Sequence(SubtreesM);
map_m_subtrees(F, Nodes, Opts) ->
    SubtreesM = m_subtrees(F, Nodes, Opts),
    monad_deep_sequence_m(SubtreesM, Opts).

m_subtrees(F, Subtrees, Opts) ->
    lists:map(
      fun({skip, Subtree}) ->
              deep_return(Subtree, Opts);
         ({Type, Subtree}) when is_atom(Type) ->
              deep_m_subtrees(F, Subtree, Opts#{node => Type});
         ({Opts1, Subtree}) when is_map(Opts1) ->
              deep_m_subtrees(F, Subtree, maps:merge(Opts, Opts1));
         (Subtree) when is_list(Subtree) ->
              deep_m_subtrees(F, Subtree, Opts)
      end, Subtrees).

deep_return(Nodes, Opts) when is_list(Nodes) ->
    lists:map(
      fun(Node) ->
              deep_return(Node, Opts)
      end, Nodes);
deep_return(Node, Opts) ->
    monad_return(Node, Opts).

deep_m_subtrees(F, Nodes, Opts) when is_list(Nodes) ->
    lists:map(
      fun(Node) ->
              deep_m_subtrees(F, Node, Opts)
      end, Nodes);

deep_m_subtrees(F, Node, Opts) ->
    map_m_1(F, Node, Opts).

%% attribute_name({tree, atom, _, Name}) ->
%%     Name.

monad_bind(A, AFB, #{monad_class := MonadClass, monad := Monad}) ->
    MonadClass:bind(A, AFB, Monad).
monad_return(A, #{monad_class := MonadClass, monad := Monad}) ->
    MonadClass:return(A, Monad).
monad_map_m(F, MAs, #{monad_class := MonadClass, monad := Monad}) ->
    MonadClass:map_m(F, MAs, Monad).
monad_lift_m(F, MAs, #{monad_class := MonadClass, monad := Monad}) ->
    MonadClass:lift_m(F, MAs, Monad).

monad_deep_sequence_m([MA|MAs], #{} = Opts) -> 
    monad_bind(
      monad_deep_sequence_m(MA, Opts),
      fun(A) ->
              monad_bind(
                monad_deep_sequence_m(MAs, Opts),
                fun(As) ->
                        monad_return([A|As], Opts)
                end, Opts)
      end, Opts);
monad_deep_sequence_m([], Opts) -> 
    monad_return([], Opts);
monad_deep_sequence_m(MA, _Opts) -> 
    MA.

monad_deep_r_sequence_m(MAs, Opts) ->
    monad_lift_m(fun lists:reverse/1, monad_deep_sequence_m(lists:reverse(MAs), Opts), Opts).

transform_mapfold_f(F, Opts) ->
    fun(Node, Attr) ->
            astranaut_traverse_monad:bind(
              astranaut_traverse_monad:get(),
              fun(State) ->
                      Reply = F(Node, State, Attr),
                      fun_return_to_monad(Reply, Node, Opts#{with_state => true})
              end)
    end.

fun_return_to_monad(Return, Node) ->
    fun_return_to_monad(Return, Node, #{}).

fun_return_to_monad(Return, Node, Opts) ->
    SyntaxLib = syntax_lib(Opts),
    MA = astranaut_traverse_monad:return(Node),
    Line = SyntaxLib:get_pos(Node),
    fun_return_to_monad_1(Return, MA, Opts#{line => Line}).

%% transform user sytle traverse return to astranaut_traverse_monad
fun_return_to_monad_1(#{'__struct__' := ?TRAVERSE_FUN_RETURN} = Struct, MA, Opts) ->
    struct_to_monad(Struct, MA, Opts);
fun_return_to_monad_1(continue, MA, _Opts) ->
    astranaut_traverse_monad:then(MA, astranaut_traverse_monad:return(continue));
fun_return_to_monad_1({continue, Node}, MA, _Opts) ->
    astranaut_traverse_monad:then(MA, astranaut_traverse_monad:return({continue, Node}));
fun_return_to_monad_1({error, Reason}, MA, Opts) ->
    NReason = format_error(Reason, Opts),
    astranaut_traverse_monad:then(MA, astranaut_traverse_monad:fail(NReason));
fun_return_to_monad_1({error, Node, Reason}, MA, Opts) ->
    NReason = format_error(Reason, Opts),
    astranaut_traverse_monad:then(
      astranaut_traverse_monad:then(MA, astranaut_traverse_monad:error(NReason)),
      astranaut_traverse_monad:return(Node));
fun_return_to_monad_1({warning, Node, Reason}, MA, Opts) ->
    NReason = format_error(Reason, Opts),
    astranaut_traverse_monad:then(
      astranaut_traverse_monad:then(MA, astranaut_traverse_monad:warning(NReason)),
      astranaut_traverse_monad:return(Node));
fun_return_to_monad_1({astranaut_monad_state_t, _} = MonadB, _MA, #{}) ->
    MonadB;
fun_return_to_monad_1({Node, State}, MA, #{with_state := true}) ->
    astranaut_traverse_monad:then(MA, astranaut_traverse_monad:state(fun(_) -> {Node, State} end));
fun_return_to_monad_1(Node, MA, _Opts) ->
    astranaut_traverse_monad:then(MA, astranaut_traverse_monad:return(Node)).

struct_to_monad(#{error := Error} = Reply, MA, Opts) ->
    NError = format_error(Error, Opts),
    MB = astranaut_traverse_monad:left_then(MA, astranaut_traverse_monad:error(NError)),
    NReply = maps:remove(error, Reply),
    struct_to_monad(NReply, MB, Opts);
struct_to_monad(#{errors := Errors} = Reply, MA, Opts) ->
    NErrors = format_errors(Errors, Opts),
    MB = astranaut_traverse_monad:left_then(MA, astranaut_traverse_monad:errors(NErrors)),
    NReply = maps:remove(errors, Reply),
    struct_to_monad(NReply, MB, Opts);
struct_to_monad(#{warning := Warning} = Reply, MA, Opts) ->
    NWarning = format_error(Warning, Opts),
    MB = astranaut_traverse_monad:left_then(MA, astranaut_traverse_monad:warning(NWarning)),
    NReply = maps:remove(warning, Reply),
    struct_to_monad(NReply, MB, Opts);
struct_to_monad(#{warnings := Warnings} = Reply, MA, Opts) ->
    NWarnings = format_errors(Warnings, Opts),
    MB = astranaut_traverse_monad:left_then(MA, astranaut_traverse_monad:warnings(NWarnings)),
    NReply = maps:remove(warnings, Reply),
    struct_to_monad(NReply, MB, Opts);
struct_to_monad(#{state := State} = Reply, MA, #{with_state := true} = Opts) ->
    MB = astranaut_traverse_monad:left_then(MA, astranaut_traverse_monad:put(State)),
    NReply = maps:remove(state, Reply),
    struct_to_monad(NReply, MB, Opts);
struct_to_monad(#{continue := true, node := Node} = Reply, MA, Opts) ->
    MB = astranaut_traverse_monad:then(MA, astranaut_traverse_monad:return({continue, Node})),
    NReply = maps:remove(continue, maps:remove(node, Reply)),
    struct_to_monad(NReply, MB, Opts);
struct_to_monad(#{continue := true} = Reply, MA, Opts) ->
    MB = astranaut_traverse_monad:then(MA, astranaut_traverse_monad:return(continue)),
    NReply = maps:remove(continue, Reply),
    struct_to_monad(NReply, MB, Opts);
struct_to_monad(#{node := Node} = Reply, MA, Opts) ->
    MB = astranaut_traverse_monad:then(MA, astranaut_traverse_monad:return(Node)),
    NReply = maps:remove(node, Reply),
    struct_to_monad(NReply, MB, Opts);
struct_to_monad(#{}, MA, _Opts) ->
    MA.

format_errors(Errors, Opts) ->
    lists:map(fun(Error) -> format_error(Error, Opts) end, Errors).

format_error(#{'__struct__' := ?TRAVERSE_ERROR, reason := Reason} = Error, #{formatter := Module, line := Line}) ->
    ErrorLine = maps:get(line, Error, Line),
    ErrorModule = maps:get(module, Error, Module),
    {ErrorLine, ErrorModule, Reason};
format_error({Line1, Error}, #{formatter := Module}) when is_integer(Line1) ->
    {Line1, Module, Error};
format_error({Line1, Module1, Error}, #{}) when is_integer(Line1), is_atom(Module1) ->
    {Line1, Module1, Error}; 
format_error(Error, #{line := Line, formatter := Module}) ->
    {Line, Module, Error}.

transform_f(F, #{traverse := pre} = Opts) ->
    fun(Node, #{step := pre  } = Attr) -> F(Node, Attr);
       (Node, #{step := leaf } = Attr) -> F(Node, Attr);
       (Node, #{step := post }) -> monad_return(Node, Opts)
    end;
transform_f(F, #{traverse := post} = Opts) ->
    fun(Node, #{step := pre  }) -> monad_return(Node, Opts);
       (Node, #{step := leaf } = Attr) -> F(Node, Attr);
       (Node, #{step := post } = Attr) -> F(Node, Attr)
    end;
transform_f(F, #{traverse := leaf} = Opts) ->
    fun(Node, #{step := pre  }) -> monad_return(Node, Opts);
       (Node, #{step := leaf } = Attr) -> F(Node, Attr);
       (Node, #{step := post }) -> monad_return(Node, Opts)
    end;
transform_f(F, _) ->
    F.

map_walk_return(F, WalkReturn) ->
    case WalkReturn of
        #{'__struct__' := ?TRAVERSE_FUN_RETURN} = Map ->
            Map;
        {error, Reason} ->
            {error, Reason};
        {ok, WalkReturn} ->
            {ok, F(WalkReturn)};
        WalkReturn ->
            F(WalkReturn)
    end.

simplify_return({ok, Reply, [], []}, true) ->
    Reply;
simplify_return(Other, _) ->
    Other.

monad_to_traverse_fun_return(Monad) ->
    monad_to_traverse_fun_return(Monad, #{}).

monad_to_traverse_fun_return(Monad, #{init := Init, with_state := true}) ->
    case astranaut_traverse_monad:run(Monad, Init) of
        {ok, Value, [], []} ->
            Value;
        {ok, {Node, State}, Errors, Warnings} ->
            traverse_fun_return(#{node => Node, state => State, errors => Errors, warnings => Warnings});
        {error, Errors, Warnings} ->
            traverse_fun_return(#{errors => Errors, warnings => Warnings})
    end;
monad_to_traverse_fun_return(Monad, #{} = Opts) ->
    Init = maps:get(init, Opts, ok),
    case astranaut_traverse_monad:run(Monad, Init) of
        {ok, {Node, _}, [], []} ->
            Node;
        {ok, {Node, _}, Errors, Warnings} ->
            traverse_fun_return(#{node => Node, errors => Errors, warnings => Warnings});
        {error, Errors, Warnings} ->
            traverse_fun_return(#{errors => Errors, warnings => Warnings})
    end.

to_parse_transform_return(Reply, Forms, Opts) when is_map(Opts) ->
    ParseTransForm = maps:get(parse_transform, Opts, false),
    case ParseTransForm of
        true ->
            File = file(Forms),
            parse_transform_return(Reply, File);
        false ->
            Reply
    end.

file(Forms) when is_list(Forms) ->
    case astranaut:attributes(file, Forms) of
        [{File, _}|_] ->
            File;
        _ ->
            ""
    end;
file(_) ->
    "".
